unit _Config;

/////////////////////////////////////////////////////////////////////
//
// Hi-Track Version 2
// Copyright (c) 2000-2002 Dmitry Liman [2:461/79]
//
// http://hi-files.by.ru
// http://hi-files.narod.ru
//
/////////////////////////////////////////////////////////////////////

interface

uses
  Objects, MyLib, FtnAddr, _Node, _Target;


const
  PID = 'Hi-Track 2.18';
  COPYRIGHT = 'Copyright (C) 2000-2003 Dmitry Liman [2:461/79]';

////////////////////////////////////////////////////////////////
//
// Config
//
////////////////////////////////////////////////////////////////

type
  PSession = ^TSession;
  TSession = record
    CfgAge: TDateTime;
    IdxAge: TDateTime;  // 0 if .nli does not exist
    BsoAge: TDateTime;
  end; { TSession }

  PConfig = ^TConfig;
  TConfig = object (TObject)
    PrimaryAddr: TAddress;
    RobotAddr  : TAddress;
    RobotName  : String;
    Netmail    : String;
    Inbound    : String;
    Outbound   : String;
    UTC        : Integer;
    LogName    : String;
    LogFile    : Text;
    LogOpened  : Boolean;
    NodelistLoaded: Boolean;

    Session    : TSession;
    SesChanged : Boolean;

    Targets  : PCollection;
    Origins  : PStrings;
    Nodelist : PNodelist;

    constructor Init;
    destructor Done; virtual;
    function Load : Boolean;
    procedure Run;
    function GetOrigin: String;
    function OpenNodelist: Boolean;
    function MakeOurViaStamp: String;
  private
    procedure OpenSession;
    procedure CloseSession;
  end; { TConfig }

type
  TLogLevel = ( ll_Start, ll_Stop, ll_Target, ll_Hook, ll_Header, ll_Action, ll_Error );

procedure Log( Level: TLogLevel; S: String );

var
  Cfg: TConfig;

{ =================================================================== }

implementation

uses
  SysUtils, _Preproc, _Catcher;

const
  DEFAULT_ORIGIN = 'Hi-Track - rulez .)';

////////////////////////////////////////////////////////////////
//
// Config
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Init                                                      }
{ --------------------------------------------------------- }

constructor TConfig.Init;
begin
  inherited Init;
  Targets  := New( PCollection, Init(10, 10) );
  Origins  := New( PStrings, Init(10, 10) );
  LogName  := ChangeFileExt( ParamStr(0), '.log' );
  Nodelist := New( PNodelist, Init );
  SetupDateTimeFormat;
end; { Init }


{ --------------------------------------------------------- }
{ Done                                                      }
{ --------------------------------------------------------- }

destructor TConfig.Done;
begin
  CloseSession;
  if LogOpened then
    Close( LogFile );
  Destroy( Targets );
  Destroy( Origins );
  Destroy( Nodelist );
  inherited Done;
end; { Done }

{ --------------------------------------------------------- }
{ Load                                                      }
{ --------------------------------------------------------- }

function TConfig.Load: Boolean;

  function NewTarget( const S: String ) : PTarget;
  begin
    Result := New( PTarget, Init );
    Targets^.Insert( Result );
    Result^.AddStorage( S );
  end; { NewTarget }

var
  j: Integer;
  S: String;
  Key : String;
  Val : String;

  Proc: TPreprocessor;

  Target : PTarget;
  Catcher: PCatcher;

  in_target: Boolean;

begin
  Result := False;

  Proc.Init;

  for j := 1 to ParamCount do
    Proc.Define( ParamStr(j), '' );

  try
    Proc.Exec( ChangeFileExt( ParamStr(0), '.cfg' ) );
  except
    on E: Exception do
      begin
        Writeln( 'Preprocessor error: ' + E.Message );
        Proc.Done;
        Exit;
      end;
  end;

  in_target := False;
  Target    := nil;
  Catcher   := nil;

  try
    while Proc.GetLine( S ) do
    begin
      SplitPair( S, Key, Val );

      if Target = nil then
      begin
        // Before the first target definition
        if cmpi( Key, 'Target' ) = 0 then
        begin
          // The first target
          in_target := True;
          Target := NewTarget( Val );
        end
        else if cmpi( Key, 'Address' ) = 0 then
          PrimaryAddr.Assign( Val )
        else if cmpi( Key, 'Robot_Addr' ) = 0 then
          RobotAddr.Assign( Val )
        else if cmpi( Key, 'Robot_Name' ) = 0 then
          RobotName := Val
        else if cmpi( Key, 'Log' ) = 0 then
          LogName := Val
        else if cmpi( Key, 'Netmail' ) = 0 then
          Netmail := Val
        else if cmpi( Key, 'Inbound' ) = 0 then
          Inbound := Val
        else if cmpi( Key, 'Outbound' ) = 0 then
          Outbound := Val
        else if cmpi( Key, 'Origin' ) = 0 then
          Origins^.Insert( AllocStr(Val) )
        else if cmpi( Key, 'Nodelist' ) = 0 then
          Nodelist^.AddFile( Val )
        else if cmpi( Key, 'UTC' ) = 0 then
          UTC := StrToInt( Val )
        else
          raise Exception.Create( 'Bad or misplaced token, line: "' + s + '"' );
      end
      else if Catcher <> nil then
      begin
        // Inside catch block
        if Catcher^.Eat( Key, Val ) then
          Catcher := nil;
      end
      else if cmpi( Key, 'Catch' ) = 0 then
      begin
        // Entering catch block
        in_target := False;
        Catcher := Target^.AddCatcher( New(PCatcher, Init) );
      end
      else if cmpi( Key, 'EasyRoute' ) = 0 then
      begin
        // Simulating catch block by easyroute 'macro' :)
        in_target := False;
        PEasyRouter(Target^.AddCatcher( New(PCatcher, Init) ))^.Simulate( Val );
      end
      else if cmpi( Key, 'Target' ) = 0 then
      begin
        if in_target then
          Target^.AddStorage( Val )
        else
        begin
          in_target := True;
          Target := NewTarget( Val );
        end
      end
      else
        raise Exception.Create( 'Bad or misplaced token, line: "' + S + '"' );
    end;

    if Origins^.Count = 0 then
      Origins^.Insert( AllocStr(DEFAULT_ORIGIN) );

    CheckDir( Netmail, 'Netmail' );
    CheckDir( Inbound, 'Inbound' );

    Assign( LogFile, LogName );
    try
      Append( LogFile );
    except
      try
        Rewrite( LogFile );
      except
        raise Exception.Create( 'Could not open/create log file "' + LogName + '"' );
      end;
    end;

    Writeln( '* Using log file "' + LogName + '"' );
    LogOpened := True;

    OpenSession;

    Result := True;

  except
    on E: Exception do
        Writeln( '! ' + E.Message );
  end;

  Proc.Done;

end; { Load }

{ --------------------------------------------------------- }
{ OpenSession                                               }
{ --------------------------------------------------------- }

type
  TSessionHeader = record
    Sign: array [0..31] of Char;
    Ver : DWORD;
  end; { TSessionHeader }

const
  HeaderSign: String[32] = 'Hi-Track session file';
  HeaderVer = 1;

procedure TConfig.OpenSession;
var
  Age   : TDateTime;
  Stream: TBufStream;
  Header: TSessionHeader;
begin
  Stream.Init( ChangeFileExt(ParamStr(0), '.ses'), stOpenRead, 512 );
  try
    if Stream.Status <> stOk then
      raise Exception.Create( 'Session file not found' );
    Stream.Read( Header, SizeOf(Header) );
    if StrLComp( @Header.Sign, @HeaderSign[1], Length(HeaderSign) ) <> 0 then
      raise Exception.Create( 'Session file corrupted' );
    if Header.Ver <> Header.Ver then
      raise Exception.Create( 'Unknown version of the session file' );
    Stream.Read( Session, SizeOf(Session) );
    if Stream.Status <> stOk then
      raise Exception.Create( 'Failed to read session file' );
  except
    on E: Exception do
      begin
        Log( ll_Error, E.Message );
        FillChar( Session, SizeOf(Session), 0 );
      end;
  end;
  Stream.Done;
  Age := FileDateToDateTime( FileAge( ChangeFileExt(ParamStr(0), '.cfg') ) );
  if Age <> Session.CfgAge then
  begin
    Session.CfgAge := Age;
    SesChanged := True;
  end;
  if not FileExists( GetNodelistIndexName ) then
    Session.IdxAge := 0;
end; { OpenSession }

{ --------------------------------------------------------- }
{ CloseSession                                              }
{ --------------------------------------------------------- }

procedure TConfig.CloseSession;
var
  Stream: TBufStream;
  Header: TSessionHeader;
begin
  if SesChanged then
  begin
    Stream.Init( ChangeFileExt(ParamStr(0), '.ses'), stCreate, 512 );
    if Stream.Status <> stOk then
      Log( ll_Error, 'Could not create session file' )
    else
    begin
      FillChar( Header, SizeOf(Header), 0 );
      Move( HeaderSign[1], Header.Sign, Length(HeaderSign) );
      Header.Ver := HeaderVer;
      Stream.Write( Header, SizeOf(Header) );
      Stream.Write( Session, SizeOf(Session) );
    end;
    Stream.Done;
  end;
end; { CloseSession }

{ --------------------------------------------------------- }
{ Run                                                       }
{ --------------------------------------------------------- }

procedure TConfig.Run;

  procedure DoShot( T: PTarget ); far;
  begin
    T^.Shot;
  end; { DoShot }

begin
  Log( ll_Start, PID + ' started' );
  Targets^.ForEach( @DoShot );
  Log( ll_Stop, PID + ' stopped' );
end; { Run }

{ --------------------------------------------------------- }
{ GetOrigin                                                 }
{ --------------------------------------------------------- }

function TConfig.GetOrigin: String;
begin
  Result := VString( Origins^.At( Random(Origins^.Count) ))^
end; { GetOrigin }

{ --------------------------------------------------------- }
{ OpenNodelist                                              }
{ --------------------------------------------------------- }

function TConfig.OpenNodelist: Boolean;
begin
  if Nodelist = nil then
    Result := False
  else if NodelistLoaded then
    Result := True
  else
  begin
    try
      Nodelist^.Open;
    except
      Destroy( Nodelist );
      Result := False;
      Exit;
    end;
    NodelistLoaded := True;
    Result := True;
  end;
end; { OpenNodelist }

{ --------------------------------------------------------- }
{ MakeOurViaStamp                                           }
{ --------------------------------------------------------- }

function TConfig.MakeOurViaStamp: String;
begin
  Result := PrimaryAddr.MakeStr + ' @' +
    FormatDateTime( 'yyyymmdd.hhmmss', Now ) + '.UTC+' + IntToStr(UTC) + ' ' + PID;
end; { MakeOurViaStamp }

{ --------------------------------------------------------- }
{ Log                                                       }
{ --------------------------------------------------------- }

type
  TLogLevelSet = set of TLogLevel;

const
  ConsoleLevels: TLogLevelSet = [ ll_Target, ll_Error ];

procedure Log( Level: TLogLevel; S: String );
var
  C: Char;
begin
  case Level of

    ll_Start:
      begin
        Writeln( Cfg.LogFile );
        C := #253;
      end;

    ll_Stop   : C := #253;
    ll_Hook   : C := #251;
    ll_Header : C := ' ';
    ll_Action : C := '+';
    ll_Error  : C := '!';
    ll_Target : C := '*';
  else
    C := '?';
  end;

  S := C + ' ' + S;

  if Level in ConsoleLevels then
    Writeln( S );

  Writeln( Cfg.LogFile, FormatDateTime('dd.mm.yyyy hh:mm:ss', SysUtils.Now), ' : ', S )

end; { Log }

end.



