unit _Node;

/////////////////////////////////////////////////////////////////////
//
// Hi-Track Version 2
// Copyright (c) 2000-2002 Dmitry Liman [2:461/79]
//
// http://hi-files.by.ru
// http://hi-files.narod.ru
//
/////////////////////////////////////////////////////////////////////

interface

uses
  Objects, MyLib, FtnAddr;

type
  TNodelistPrefix = ( np_none, np_zone, np_region, np_host,
                      np_hub, np_pvt, np_hold, np_down);

  PNode = ^TNode;
  TNode = packed record
    Node: SWORD;
    Net : SWORD;
    HubNode: SWORD;
    HubNet : SWORD;
    Prefix : Byte;
  end; { TNode }

  PNodelist = ^TNodelist;
  TNodelist = object (TSortedCollection)
    constructor Init;
    destructor Done; virtual;
    procedure FreeItem( Item: Pointer ); virtual;
    function Compare( Key1, Key2: Pointer ) : Integer; virtual;
    procedure PutItem( var S: TStream; Item: Pointer ); virtual;
    function GetItem( var S: TStream ) : Pointer; virtual;
    procedure AddFile( const Pattern: String );
    procedure Open;
    function  UpToDate: Boolean;
    procedure Compile;
    function FindHub( A: TAddress; var Hub: TAddress ) : Boolean;
  private
    Files: PStrings;
  end; { TNodelist }

function GetNodelistIndexName: String;

{ =================================================================== }

implementation

uses SysUtils, _Config, _MapFile;

{ --------------------------------------------------------- }
{ GetNodelistIndexName                                      }
{ --------------------------------------------------------- }

const
  INDEX_EXT = '.nli';

function GetNodelistIndexName: String;
begin
  Result := ChangeFileExt( ParamStr(0), INDEX_EXT );
end; { GetNodelistIndexName }

////////////////////////////////////////////////////////////////
//
// TNodelist
//
////////////////////////////////////////////////////////////////

const
  STREAM_BUFFER_SIZE = 32 * 1024;
  CURRENT_VERSION = 1;

const
  HeaderSign: String[32] = 'Hi-Track nodelist index';

type
  PNodelistHeader = ^TNodelistHeader;
  TNodelistHeader = record
    Sign  : array [0..31] of Char;
    Ver   : DWORD;
    Count : DWORD;
  end; { TNodelistHeader }

{ --------------------------------------------------------- }
{ Init                                                      }
{ --------------------------------------------------------- }

constructor TNodelist.Init;
begin
  inherited Init(1000, 1000);
  Files := New( PStrings, Init(10, 10) );
end; { Init }

{ --------------------------------------------------------- }
{ Done                                                      }
{ --------------------------------------------------------- }

destructor TNodelist.Done;
begin
  Destroy( Files );
  inherited Done;
end; { Done }

{ --------------------------------------------------------- }
{ FreeItem                                                  }
{ --------------------------------------------------------- }

procedure TNodelist.FreeItem( Item: Pointer );
begin
  Dispose( PNode(Item) );
end; { FreeItem }

{ --------------------------------------------------------- }
{ Compare                                                   }
{ --------------------------------------------------------- }

function TNodelist.Compare( Key1, Key2: Pointer ) : Integer;
var
  n1: PNode absolute Key1;
  n2: PNode absolute Key2;
begin
  if n1^.Net < n2^.Net then
    Result := -1
  else if n1^.Net > n2^.Net then
    Result := 1
  else if n1^.Node < n2^.Node then
    Result := -1
  else if n1^.Node > n2^.Node then
    Result := 1
  else
    Result := 0;
end; { Compare }

{ --------------------------------------------------------- }
{ PutItem                                                   }
{ --------------------------------------------------------- }

procedure TNodelist.PutItem( var S: TStream; Item: Pointer );
var
  n: PNode absolute Item;
begin
  S.Write( n^, SizeOf(TNode) );
end; { PutItem }

{ --------------------------------------------------------- }
{ GetItem                                                   }
{ --------------------------------------------------------- }

function TNodelist.GetItem( var S: TStream ) : Pointer;
var
  Node: PNode;
begin
  New( Node );
  S.Read( Node^, SizeOf(TNode) );
  Result := Node;
end; { GetItem }

{ --------------------------------------------------------- }
{ AddFile                                                   }
{ --------------------------------------------------------- }

procedure TNodelist.AddFile( const Pattern: String );
var
  SR: TSearchRec;
  DosError  : Integer;
  LatestFile: String;
  LatestTime: TDateTime;
  ThisTime  : TDateTime;
begin
  LatestTime := 0;
  DosError := SysUtils.FindFirst( Pattern, faArchive or faReadOnly, SR );
  while DosError = 0 do
  begin
    ThisTime := FileDateToDateTime( SR.Time );
    if ThisTime > LatestTime then
    begin
      LatestTime := ThisTime;
      LatestFile := SR.Name;
    end;
    DosError := SysUtils.FindNext( SR );
  end;
  SysUtils.FindClose( SR );

  if LatestTime > 0 then
    Files^.Insert( AllocStr(AtPath(LatestFile, ExtractFilePath(Pattern)) ));
end; { AddFile }

{ --------------------------------------------------------- }
{ Open                                                      }
{ --------------------------------------------------------- }

procedure TNodelist.Open;
var
  Bin: TBufStream;
  Header: TNodelistHeader;
begin
  if not UpToDate then
    Compile
  else
  begin
    Log( ll_Target, 'Loading nodelist index...' );
    Bin.Init( GetNodelistIndexname, stOpenRead, STREAM_BUFFER_SIZE );
    try
      if Bin.Status <> stOk then
        raise EInOutError.Create( 'Could not open nodelist index: ' + GetNodelistIndexName );

      Bin.Read( Header, SizeOf(Header) );
      if StrLComp( @Header.Sign, @HeaderSign[1], Length(HeaderSign) ) <> 0 then
        raise Exception.Create( 'Nodelist index file corrupted' );

      if Header.Ver <> CURRENT_VERSION then
        raise Exception.Create( 'Unsupported version of nodelist index file (0x' +
                                IntToHex(Header.Ver, 8) + ')' );

      Load( bin );

    finally
      Bin.Done;
    end;
  end;
end; { Open }

{ --------------------------------------------------------- }
{ UpToDate                                                  }
{ --------------------------------------------------------- }

function TNodelist.UpToDate: Boolean;
var
  IndexAge: TDateTime;

  function Updated( V: VString ) : Boolean; far;
  begin
    if FileAge( V^ ) > IndexAge then
    begin
      Log( ll_Target, 'Updated nodelist detected: ' + V^ );
      Result := True;
    end
    else
      Result := False;
  end; { Updated }

begin
  Result := False;
  IndexAge := FileAge( GetNodelistIndexName );
  if IndexAge < 0 then
  begin
    Log( ll_Target, 'Nodelist index not found' );
    Exit;
  end;
  Result := Files^.FirstThat( @Updated ) = nil;
end; { UpToDate }

{ --------------------------------------------------------- }
{ Compile                                                   }
{ --------------------------------------------------------- }

procedure TNodelist.Compile;

  procedure ParseLine( const S: String; var Prefix: String; var Number: Integer );
  begin
    if S[1] = ',' then
    begin
      Prefix := '';
      Number := StrToInt( ExtractWord(1, S, _COMMA) );
    end
    else
    begin
      Prefix := ExtractWord( 1, S, _COMMA );
      Number := StrToInt( ExtractWord( 2, S, _COMMA ));
    end;
  end; { ParseLine }

  procedure DoCompile( FileName: VString ); far;
  var
    S: String;
    Map: TMappedFile;
    Node    : PNode;
    Host    : Integer;
    Prefix  : String;
    Number  : Integer;
    HubNode : Integer;
    HubNet  : Integer;
  begin
    Log( ll_Target, 'Compiling ' + Filename^ );
    Host := 0;
    HubNode := 0;
    HubNet  := 0;
    Map.Init( FileName^ );
    while Map.GetLine( S ) do
    begin
      if (S = '') or (S[1] = ';') or (S[1] = ^Z) then
        Continue;

      try
        ParseLine( S, Prefix, Number );
      except
        Log( ll_Error, 'Bad nodelist line ignored: ' + S );
        Continue;
      end;

      New( Node );
      FillChar( Node^, SizeOf(TNode), 0 );
      Node^.Node := Number;

      if Prefix = '' then
        Node^.Prefix := Ord(np_none)
      else if Prefix = 'Hub' then
      begin
        Node^.Prefix := Ord(np_hub);
        HubNode := Number;
      end
      else if Prefix = 'Pvt' then
        Node^.Prefix := Ord(np_pvt)
      else if Prefix = 'Hold' then
        Node^.Prefix := Ord(np_hold)
      else if Prefix = 'Down' then
        Node^.Prefix := Ord(np_down)
      else if (Prefix = 'Host') or (Prefix = 'Region') then
      begin
        if Prefix = 'Host' then
          Node^.Prefix := Ord(np_host)
        else
          Node^.Prefix := Ord(np_region);

        Host := Number;

        Node^.Node := 0;
        Node^.Net  := Host;
        HubNode := 0;
        HubNet  := Host;
      end
      else if Prefix = 'Zone' then
      begin
        // Multi-zone nodelist is not supported yet
        Dispose( Node );
        Continue;
      end
      else
      begin
        Log( ll_Error, 'Unknown nodelist line ignored: "' + S + '"' );
        Dispose( Node );
        Continue;
      end;

      Node^.Net     := Host;
      Node^.HubNet  := HubNet;
      Node^.HubNode := HubNode;
      Insert( Node );
    end;
  end; { DoCompile }

var
  Header: TNodelistHeader;
  Bin   : TBufStream;

begin
  if Files^.Count = 0 then
    Log( ll_Error, 'No nodelist files specified, routing features will reduced' )
  else
  begin
    Log( ll_Target, 'Building nodelist index...' );
    Files^.ForEach( @DoCompile );
  end;

  FillChar( Header, SizeOf(Header), 0 );
  Move( HeaderSign[1], Header.Sign, Length(HeaderSign) );
  Header.Ver := CURRENT_VERSION;
  Header.Count := Count;

  Bin.Init( GetNodelistIndexName, stCreate, STREAM_BUFFER_SIZE );
  if Bin.Status <> stOk then
  begin
    Bin.Done;
    raise EInOutError.Create( 'Could not create nodelist index: ' + GetNodelistIndexName );
  end;

  Bin.Write( Header, SizeOf(Header) );
  Store( Bin );
  Bin.Done;

  Cfg.Session.IdxAge := SysUtils.Now;
  Cfg.SesChanged := True;
end; { Compile }

{ --------------------------------------------------------- }
{ FindHub                                                   }
{ --------------------------------------------------------- }

function TNodelist.FindHub( A: TAddress; var Hub: TAddress ) : Boolean;
var
  N: TNode;
  H: PNode;
  j: Integer;
begin
  N.Node := A.Node;
  N.Net  := A.Net;
  if Search( @N, j ) then
  begin
    H := At( j );
    Hub.Zone  := A.Zone;
    Hub.Net   := H^.HubNet;
    Hub.Node  := H^.HubNode;
    Hub.Point := 0;
    Result := True;
  end
  else
    Result := False;
end; { FindHub }

end.

